/*
 *  Cryptographic "Data-as-Key" Feedback Analyser
 *
 *  Copyright (C) 2005 Luke Kenneth Casson Leighton
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*#define TEST_HISTOGRAM*/

#ifdef TEST_RIJNDAEL
#include "aes.h"
#define ALG "rjd"
#endif

#ifdef TEST_SERPENT
#include "serpent.h"
#define ALG "serpent"
#endif

#ifdef TEST_BF
#include <openssl/blowfish.h>
#define ALG "blowfisn"
#endif

#ifdef TEST_CAST
#include <openssl/cast.h>
#define ALG "cast"
#endif

#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "tomcrypt.h"

#define sqrt2 1.41421356237309504880
/*
 * the basic premise of this program is as follows:
 *
 * for a given initial key "k" and initial data "buf",
 * repeat num_tests times:
 *
 * 	o = E(k, buf); k = o; buf = xor(buf, o)
 *
 * if the algorithm is any good, then using the key as the
 * data for the next block will show no signs of non-randomness
 * in the output.  this just seems like good common sense.
 *
 * (and Rijndael/128 shows problems on blocks as quickly as 5-10 loops)
 *
 * signs of non-randomness aren't such a big hairy deal... as long as
 * there's no correlation between bits being set in the initial key
 * and bits being set in the initial data... and that's what the
 * histogram is there to find out.
 *
 * use gnuplot in weird and wonderful ways to look at the plots in 3d...
 *
 */

static double pval_bitbucket[128][128];

extern int num_bits(unsigned char *data, int len);
extern double erfc(double x);

void bit_set(unsigned char *d, int bit)
{
	d[bit/8] |= 1<<(bit&0x7);
}

int bit_is_set(unsigned char *d, int bit)
{
	return (d[bit/8] & 1<<(bit&0x7)) != 0;
}

static void xor(unsigned char *key2, const unsigned char *key1,
				const unsigned char *xorval)
{
	int i;
	for (i = 0; i < 16; i++)
		key2[i] = key1[i] ^ xorval[i];
}

static void encrypt(const unsigned char *key, const unsigned char *buf,
					unsigned char *out)
{
	/* o = E(k, buf); k = o; buf = xor(buf, o) */
#ifdef TEST_RIJNDAEL
	aes_context ctx;
	aes_set_key( &ctx, (unsigned char*)key, 128 );
	aes_encrypt( &ctx, (unsigned char*)buf, out );
#endif
#ifdef TEST_SERPENT
	unsigned int ctx[140];
	serpent_set_key( ctx, key, 128 );
	serpent_encrypt( &ctx, buf, out);
#endif
#ifdef TEST_CAST
	CAST_KEY ctx;
	CAST_set_key( &ctx, 16, key );
	CAST_ecb_encrypt( &buf[0], &out[0], &ctx, CAST_ENCRYPT);
	CAST_ecb_encrypt( &buf[8], &out[8], &ctx, CAST_ENCRYPT);
#endif
#ifdef TEST_BF
	BF_KEY ctx;
	BF_set_key( &ctx, 16, key );
	BF_ecb_encrypt( &buf[0], &out[0], &ctx, BF_ENCRYPT);
	BF_ecb_encrypt( &buf[8], &out[8], &ctx, BF_ENCRYPT);
#endif
}

double data_test(unsigned char *test_data, int len)
{
	int count;
	double sobs;
	double testval;
	double f;

	/* count the number of bits, create a p-value, blah blah */
	count = num_bits(test_data, len);
	testval = (count - ((double)len*4)) * 2;
	sobs = fabs(testval) / sqrtl(len*8);
	f = sobs / sqrt2;
	return erfc(f);
}

/* 
 * o1 = E(k xor xorkey, d xor xordata)
 * o2 = E(k           , d)
 * o = o1 xor o2
 *
 * count num bits in o
 *
 */
int analyse(  unsigned char *baseline,
		unsigned char *d, unsigned char *k, unsigned char *xordata,
		unsigned char *xorkey,
		unsigned char *out)
{
	unsigned char buf[16];
	/*unsigned char out[16];*/
	unsigned char key[16];
	unsigned char test_data[16]; /* make this 32..64... for larger tests */
	double pval;

	int num_fails = 0;

	memcpy(key, k, 16);
	memcpy(buf, d, 16);

	xor(key, key, xorkey);
	xor(buf, buf, xordata);
	encrypt(key, buf, out);

	xor(test_data, out, baseline);

	/* count the number of bits, create a p-value */
	pval = data_test(test_data, sizeof(test_data));

	if (pval < 0.01)
		num_fails++;


	/*printf("testval: %10.f pval: %f\n", testval, pval);*/
	return num_fails;
}

/* okay, this is really simple.
 * repeat num_tests times:
 *
 * 	o = E(k, buf); k = o; buf = xor(buf, o)
 *
 */
int analyse2( unsigned char *d, unsigned char *k, unsigned char *xordata,
		int num_tests)
{
	int i, j;
	unsigned char buf[16];
	unsigned char out[16];
	unsigned char key[16];
	unsigned char test_data[16]; /* make this 32..64... for larger tests */
	int count;
	double sobs;
	double testval;
	double pval;
	double f;

	int num_fails = 0;

	memcpy(key, k, 16);
	memcpy(buf, d, 16);

	for (i = 0; i < num_tests; i++)
	{
		for (j = 0; j < sizeof(test_data); j += 16)
		{
			/* o = E(k, buf); k = o; buf = xor(buf, o) */
#ifdef TEST_RIJNDAEL
			aes_context ctx;
			aes_set_key( &ctx, key, 128 );
			aes_encrypt( &ctx, buf, out );
#endif
#ifdef TEST_SERPENT
			unsigned int ctx[140];
			serpent_set_key( ctx, key, 128 );
			serpent_encrypt( &ctx, buf, out);
#endif
#ifdef TEST_CAST
			CAST_KEY ctx;
			CAST_set_key( &ctx, 16, key );
			CAST_ecb_encrypt( &buf[0], &out[0], &ctx, CAST_ENCRYPT);
			CAST_ecb_encrypt( &buf[8], &out[8], &ctx, CAST_ENCRYPT);
#endif
#ifdef TEST_BF
			BF_KEY ctx;
			BF_set_key( &ctx, 16, key );
			BF_ecb_encrypt( &buf[0], &out[0], &ctx, BF_ENCRYPT);
			BF_ecb_encrypt( &buf[8], &out[8], &ctx, BF_ENCRYPT);
#endif
			memcpy(key, out, 16);
			xor(buf, out, buf);
			xor(buf, buf, xordata);

			/* output - concatenate data for testing... */
			memcpy(&test_data[j], out, 16);
		}

		/* count the number of bits, create a p-value, blah blah */
		count = num_bits(test_data, sizeof(test_data));
		testval = (count - ((double)sizeof(test_data)*4)) * 2;
		sobs = fabs(testval) / sqrtl(sizeof(test_data));
		f = sobs / sqrt2;
		pval = erfc(f);

		/*if (pval < 0.01)*/
		if (pval > 0.99)
			num_fails++;
		/*printf("testval: %10.f pval: %f\n", testval, pval);*/
	}
	return num_fails;
}

struct datapoint
{
	unsigned char i;
	unsigned char j;
	int bitbucket;
	struct datapoint *next;
};

static int stats[128][128][128][128];
static int bitbucket[128][128];
static struct datapoint *sortedbucket = NULL;

void dump_bitbucket(void)
{
	fwrite(bitbucket, sizeof(bitbucket), 1, stdout);
}
			

void print_bitbucket(int num_samples)
{
	int i, j;
	for (i = 0; i < 128; i++)
	{
		int print = 0;
		for (j = 0; j < 128; j++)
		{
			if (bitbucket[i][j] != 0)
			{
				print = 1;
				break;
			}
		}
		if (!print)
			continue;

		for (j = 0; j < 128; j++)
		{
			if (bitbucket[i][j] != 0)
				printf("%3d ", (int)(100.*128.*128.*(double)bitbucket[i][j]/(double)num_samples));
		}
		printf("\n");
	}
}


int test(struct datapoint *i1, struct datapoint *i2)
{
	return 1;
	if (i1->bitbucket > i2->bitbucket)
		return 1;
	/*if (i1->bitbucket == i2->bitbucket && i1->j > i2->j)
		return 1;*/
	return 0;
}

void insert_datapoint(int i, int j, int bitbucket)
{
	struct datapoint *d = malloc(sizeof(struct datapoint));
	struct datapoint *start = sortedbucket;

	d->i = i;
	d->j = j;
	d->bitbucket = bitbucket;
	d->next = NULL;

	if (start == NULL)
	{
		sortedbucket = d;
		return;
	}

	if (test(sortedbucket, d))
	{
		d->next = sortedbucket;
		sortedbucket = d;
		return;
	}

	while (start->next != NULL)
	{
		if (test(start->next, d))
		{
			d->next = start->next;
			start->next = d;
			return;
		}
		start = start->next;
	}
	start->next = d;
}

static unsigned char rnumsj[128];
static unsigned char rnumsi[128];

static void remove_old_graph(const char *alg_name, int samples)
{
	char fname[256];

	sprintf(fname, "%s.%d.txt", alg_name, samples);
	unlink(fname);
}

static void write_floatbucket(double bb[128][128],
		const char *alg_name, int samples)
{
	int i, j;
	char fname[256];
	FILE *f;

	sprintf(fname, "%s.%d.txt", alg_name, samples);
	f = fopen(fname, "w");

	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			fprintf(f, "%d.\t%d.\t%g.\n", i, j, bb[i][j]);
		}
	}
	fclose(f);
}

static void write_bucket(int bb[128][128],
		const char *alg_name, int samples)
{
	int i, j;
	char fname[256];
	FILE *f;

	sprintf(fname, "%s.%d.txt", alg_name, samples);
	f = fopen(fname, "w");

	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			fprintf(f, "%d.\t%d.\t%d.\n", i, j, bb[i][j]);
		}
	}
	fclose(f);
}

static void print_sortedbucket(const char *alg_name, int samples)
{
	write_bucket(bitbucket, alg_name, samples);
}

void swap(unsigned char *rnums, unsigned char i, unsigned char j)
{
	unsigned char tmp = rnums[i];
	rnums[i] = rnums[j];
	rnums[j] = tmp;
}

void init_rnums(unsigned char *rnums)
{
	unsigned char i;
	for (i = 0; i < 128; i++)
		rnums[i] = i;
}

void init_ij_rnums(void)
{
	init_rnums(rnumsj);
	init_rnums(rnumsi);
}

void print_ij_rnums(void)
{
	unsigned char i;
	for (i = 0; i < 128; i++)
		printf("%3d %3d\n", rnumsi[i], rnumsj[i]);
}

void diffcount( int *icount, int *jcount)
{
	struct datapoint *d;
	d = sortedbucket;
	*jcount = 0;
	*icount = 0;
	while (d != NULL)
	{
		int idiff = abs(d->j - rnumsi[d->i]);
		int jdiff = abs(d->i - rnumsj[d->j]);

		*icount += idiff;
		*jcount += jdiff;

		d = d->next;
	}
}

#if 0

static int prevdiffi = 1<<30;
static int prevdiffj = 1<<30;

void get_best_diffcount(void)
{
	int counti = 0;
	int countj = 0;

	int r = rand();

	int i_j = (r>>0) & 127;
	int i_i = (r>>7) & 127;
	int j_j = (r>>14) & 127;
	int j_i = (r>>21) & 127;

	static int rnums_inited;
	
	if (!rnums_inited)
	{
		init_ij_rnums();
		rnums_inited = 1;
	}

	swap(rnumsi, i_i, i_j);
	swap(rnumsj, j_i, j_j);

	diffcount(&counti, &countj);

	if (countj >= prevdiffj)
	{
		swap(rnumsj, j_i, j_j);
	}
	else
	{
		prevdiffj = countj;
	}
	if (counti >= prevdiffi)
	{
		swap(rnumsi, i_i, i_j);
	}
	else
	{
		prevdiffi = counti;
	}

}

void repeat_best_count(int goes)
{
	int i;
	for (i = 0; i < goes; i++)
		get_best_diffcount();
}
#endif

extern int gen_rand_impl(unsigned char *, unsigned int len);

static void pseud(unsigned char *d)
{
	gen_rand_impl(d, 16);
}

#define RECORD_OUT 2560

int main( void )
{
	int k, i, j;
	unsigned char xorbuf[16];
	unsigned char key[16];
	unsigned char xorkey[16];
	unsigned char input[16];
	unsigned char out[16];
	unsigned char baseline[16];
	char fn[128];

	int total = 0;
	int max_run = 256000;

	int i1;

	int key_set[128];
	int bl_set[128];
	int out_set[128];

	sprintf(fn, "%s-pval", ALG);

	for (i = 0; i < 128; i++)
		for (j = 0; j < 128; j++)
			pval_bitbucket[i][j] = 0.0;

#if 0
	for (i = 0; i < 128; i++)
	{
		/* count the number of bits, create a p-value, blah blah */
		int count = i;
		double testval = (count - ((double)16*4)) * 2;
		double sobs = fabs(testval) / sqrtl(128);
		double f = sobs / sqrt2;
		double pval = erfc(f);
		/*p_buckets[i] = pval;*/

		printf("%.30f\n", pval);
	}
	exit(0);
#endif

	chdir("graphs");

	for (k = 0; k < max_run; k++) /* the more, the merrier */
	{
		pseud(key);  /*paranoia: get a decent PRNG, dude */
		pseud(input);  /*paranoia: get a decent PRNG, dude */
#if 0
		memset( input, 0x0, 16 );
		/*bit_set(input, k);*/
		(*(int*)input) = k;
#endif

		encrypt(key, input, baseline);

		for (i1 = 0; i1 < 128; i1++)
		{
			key_set[i1] = bit_is_set(key, i1);
			bl_set[i1] = bit_is_set(baseline, i1);
		}

		for (i = 0; i < 128; i++)
		{
			memset( xorbuf, 0x0, 16 );
			bit_set(xorbuf, i);
			/*bit_set(buf, k % 128);*/

			/*printf("%d-> ", i);*/
			for (j = 0; j < 128; j++)
			{
				int num_fails;
				memset( xorkey, 0x0, 16 );
				bit_set(xorkey, j);

				/* get the number of times that the
				 * p-value (see NIST SP800-22b.pdf)
				 * fell belowout2 0.01 */
				num_fails = analyse(baseline, input, key, xorbuf, xorkey, out);
				/*printf("i: %d j: %d num_fails: %d/%d\n",
						i, j, num_fails, attempts);*/

				/* create a histogram of the failures, using
				 * the key bit index and the data bit index */
				bitbucket[i][j] += num_fails;
				total += num_fails;

				if (num_fails)
				{
					double pval_out;
					unsigned char xorkd[16];
					xor(xorkd, out, key);
					pval_out = data_test(xorkd, sizeof(xorkd));

					if (pval_out < 0.005 || pval_out > 0.72)
						pval_bitbucket[i][j] += 1;
				}
				
				if (0)
				{
					int i1;
					int j1;

					if (i >= 4)
						continue;
					if (j >= 4)
						continue;

					for (i1 = 0; i1 < 128; i1++)
					{
						out_set[i1] = bit_is_set(out, i1);
					}

					for (i1 = 0; i1 < 128; i1++)
					{
						for (j1 = 0; j1 < 128; j1++)
						{
							if ((key_set[i1] ^ out_set[j1]) ==
							    (bl_set[i1]))
						/*
							if ((key_set[i1] ^ out_set[j1]) ==
							    (key_set[i1] ^ bl_set[j1]))
								*/
							{
								stats[i][j][i1][j1] += 1;
							}
						}
					}
				}
			}
		}
		fprintf(stderr, "%d\n", k);
		if ((k % RECORD_OUT == 0) && (k != 0))
		{
			int oldk = k-RECORD_OUT*2;
			if (oldk > 0)
				remove_old_graph(ALG, oldk);

			print_sortedbucket(ALG, k);

			sprintf(fn, "%s-pval", ALG);
			if (oldk > 0)
				remove_old_graph(fn, oldk);
			write_floatbucket(pval_bitbucket, fn, k);

#if 0
			for (i1 = 0; i1 < 4; i1++)
			{
				for (j1 = 0; j1 < 4; j1++)
				{
					char fn[128];
					int i2;
					int j2;
					int st[128][128];

					for (i2 = 0; i2 < 128; i2++)
						for (j2 = 0; j2 < 128; j2++)
							st[i2][j2] = stats[i1][j1][i2][j2];

					sprintf(fn, "%s-%d.%d", ALG, i1, j1);
					/*if (oldk > 0)
						remove_old_graph(fn, oldk);*/
					write_bucket(st, fn, max_run);

				}
			}
#endif
		}
	}
	/*print_bitbucket(total);*/
	/*print_sortedbucket(total);*/
	/*dump_bitbucket();*/
	print_sortedbucket(ALG, k);
	write_floatbucket(pval_bitbucket, fn, k);
#if 0
	repeat_best_count(10000);
	print_sortedbucket(total);
	print_ij_rnums();
#endif
	return( 0 );
}

